////////////////////////////////////////////////////////////////////////////
//
// Copyright 1993-2014 NVIDIA Corporation.  All rights reserved.
//
// Please refer to the NVIDIA end user license agreement (EULA) associated
// with this source code for terms and conditions that govern your use of
// this software. Any use, reproduction, disclosure, or distribution of
// this software and related documentation outside the terms of the EULA
// is strictly prohibited.
//
////////////////////////////////////////////////////////////////////////////

#ifndef _IMAGE_INTEGRAL_SUM_KERNEL_CH_
#define _IMAGE_INTEGRAL_SUM_KERNEL_CH_

#include <helper_math.h>
#include <helper_functions.h>

//RGBA float values of the image.
float4 *d_RGB_Temp, *d_RGB_Final;
float4 *h_RGB_Image, *h_RGB_Temp;
float4 offset;
float4 imageMeanValue;


////////////////////////////////////////////////////////////////////////////////
// These are CUDA Helper functions

// This will output the proper CUDA error strings in the event that a CUDA host call returns an error
#define checkCudaErrors(err)           __checkCudaErrors (err, __FILE__, __LINE__)

inline void __checkCudaErrors(cudaError err, const char *file, const int line)
{
    if (cudaSuccess != err)
    {
        fprintf(stderr, "%s(%i) : CUDA Runtime API error %d: %s.\n",
                file, line, (int)err, cudaGetErrorString(err));
        exit(EXIT_FAILURE);
    }
}



// RGBA version
// reads from 32-bit unsigned int array holding 8-bit RGBA

// convert floating point rgba color to 32-bit integer
__device__ unsigned int rgbaFloatToInt(float4 rgba)
{
    rgba.x = __saturatef(rgba.x);   // clamp to [0.0, 1.0]
    rgba.y = __saturatef(rgba.y);
    rgba.z = __saturatef(rgba.z);
    rgba.w = __saturatef(rgba.w);
    return ((unsigned int)(rgba.w * 255.0f) << 24) |
           ((unsigned int)(rgba.z * 255.0f) << 16) |
           ((unsigned int)(rgba.y * 255.0f) <<  8) |
           ((unsigned int)(rgba.x * 255.0f));
}


// convert  32-bit integer to floating point value
__host__ __device__ float4 rgbaIntToFloat(unsigned int c)
{
    float4 rgba;
    rgba.x = (c & 0xff) * 0.003921568627f;       //  /255.0f;
    rgba.y = ((c>>8) & 0xff) * 0.003921568627f;  //  /255.0f;
    rgba.z = ((c>>16) & 0xff) * 0.003921568627f; //  /255.0f;
    rgba.w = ((c>>24) & 0xff) * 0.003921568627f; //  /255.0f;
    return rgba;
}


//compute the float mean values of r g b a components of the image
__host__ float4 computeImageMeanValue( float4* imageRGBAValues, int width, int height)
{
	unsigned int imageSize = width * height;
	float4 meanValue;
	meanValue.x = 0;
	meanValue.y = 0;
	meanValue.z = 0;
	meanValue.w = 0;
	for(int i = 0; i < imageSize; i++)
	{
		meanValue += imageRGBAValues[i];
	}
	meanValue /= imageSize;
	return meanValue;
}

/*
     Perform prefix sum operation 1 row of elements

    Parameters:
	imageBuffer - pointer to the floating point values of r g b a componets buffer
	finalImage - pointer to the floating point values of r g b a ( intially ( i.e. before kernel call) consists of the r g b a floating point values of the original image)
    w  - image width
    h - image height
	row - row index whose prefix sum needs to be computed

*/
__global__ void
d_integralsum_rgba_x(float4 *imageBuffer, float4 *finalImage, int w, int h, int row)
{
  
  //TODO Implement the kernel for the prefix sum of row elements
}


/*
    Perform prefix sum operation 1 column of elements

    Parameters:
	imageBuffer - pointer to the floating point values of r g b a componets buffer
	finalImage - pointer to the floating point values of r g b a ( intially ( i.e. before kernel call) consists of the r g b a floating point values of the original image)
	destinationImage - pointer to unsigned int destination image in device memory
    w  - image width
    h - image height
	column - column index whose prefix sum needs to be computed

*/

__global__ void
d_integralsum_rgba_y(float4 *imageBuffer, float4 *finalImage, int w, int h, int column)
{
  
	//TODO Implement the kernel for the prefix sum of column elements
}


/*
    converts the integral sum image values into values which can be visualized

    Parameters:
	finalImage - pointer to the floating point values of r g b a ( intially ( i.e. before kernel call) consists of the r g b a floating point values of the original image)
	destinationImage - pointer to unsigned int destination image in device memory
    w  - image width
    h - image height
	totalSum - is the total sum of each float component ( i.e. r g b a) of the entire image  

*/
__global__ void d_assign_final_image(float4 *finalImage, unsigned int *destinationImage, int w, int h, float4 totalSum)
{
	uint i = blockDim.x * blockIdx.x + threadIdx.x;
	uint j = blockDim.y * blockIdx.y + threadIdx.y;
	uint id = j * w + i;
	finalImage[id].x = finalImage[id].x / totalSum.x;
	finalImage[id].y = finalImage[id].y / totalSum.y;
	finalImage[id].z = finalImage[id].z / totalSum.z;
	finalImage[id].w = finalImage[id].w;
	destinationImage[id] = rgbaFloatToInt(finalImage[id]);
}

/*
    Intialize image buffers for floating point rgba values (device buffers d_RGB_Temp and d_RGB_Image and host buffer h_RGB_Image)

    Parameters:
	pImage - pointer to unsigned int source image in host memory
    width  - image width
    height - image height
	

*/
extern "C"
void initRGBImageMemory(int width, int height, unsigned int *pImage, bool useRGBA)
{
    
	//allocate the device memory for storing 4 float components of image (i.e. r, g, b, a)
	cudaMalloc((void **) &d_RGB_Temp, (width * height * sizeof(float4)));
	cudaMalloc((void **) &d_RGB_Final, (width * height * sizeof(float4)));

	//allocate the host memory for storing 4 float components of image (i.e. r, g, b, a)
	h_RGB_Image = (float4 *)malloc(width * height * sizeof(float4));
	h_RGB_Temp = (float4 *)malloc(width * height * sizeof(float4));

	//convert the unsigned int from pImage to float4 and store the values in h_RGB_Image
	for (int i = 0; i < width*height; i++)
	{
		h_RGB_Image[i] = rgbaIntToFloat(pImage[i]);
		
	}

	//compute the mean values of the image.
	imageMeanValue = computeImageMeanValue(h_RGB_Image, width, height);

	//copy the above converted float r g b a values to device memory
	cudaMemcpy(d_RGB_Temp, h_RGB_Image, (width * height * sizeof(float4)), cudaMemcpyHostToDevice);
	cudaMemcpy(d_RGB_Final, h_RGB_Image, (width * height * sizeof(float4)), cudaMemcpyHostToDevice);


}

extern "C"
void freeRGBImageMemory()
{
    cudaFree(d_RGB_Temp);
    cudaFree(d_RGB_Final);
	 free(h_RGB_Image);
	 free(h_RGB_Temp);
}



/*
    Perform integral sum operation using high precision method

    Parameters:
    d_dest - pointer to unsigned int destination image in device memory
    width  - image width
    height - image height

*/
extern "C"
void integralSum(unsigned int *d_dest, int width, int height)
{


	//TODO by the student

	//2. implement and call appropriate kernels for computing the integral sum of the image using high prescision method as described in lecture
	//( whose float rgba values are stored in the device buffer "d_RGB_Temp" ( of type float4*)) and store integral sum values to "d_RGB_Final"
	// convert float values in "d_RGB_Final" into unsigned integer values and store them into "d_dest" 
	//using kernel "d_assign_final_image" however you have to assign approriate value to the parameter totalSum in order to visualize all the pixels of image
	// please refer the example below.

	// in float4 type component x = red float value, component y = green float value, component z = blue float value and component w = alpha float value (where w is always 0 in our case) 


	//PS: "h_RGB_Image" ( type float4)  is host memory consisting of float values for r g b a of the image if it is required in case
	//( each element im "h_RGB_Image" consists of r g b a float values of pixel of the image).
	

	
////------------------------------------------------Example Framework for Naive Approach-----------------------------------------------------------------------------------------/////	
//Below is an example framework of computation of image integral sum using naive approach presented in the lecture and converting it to appropriate values for visualization.

/*
	 // sync host and start computation
    checkCudaErrors(cudaDeviceSynchronize());


	//block and grid parameters for row prefix sum computation
	dim3 grid(1);
	dim3 blockRow(width);

	//Prefix sum of row elements
	for(int row = 0; row < height; row++)
	{

		d_integralsum_rgba_x<<<grid,blockRow>>>(d_RGB_Temp, d_RGB_Final, width, height, row);  

	}

	//block parameter for columns prefix sum computation
	dim3 blockColumn(height);


	//Prefix Sum of column elements
	for(int column = 0; column < width; column++)
	{

		d_integralsum_rgba_y<<<grid,blockColumn>>>(d_RGB_Temp, d_RGB_Final, width, height, column);  

	}


	dim3 blockSize(32,32);
	dim3 gridSize(width/blockSize.x, height/blockSize.y);

	//copy the image integral sum values from d_RGB_Final to host memory
	cudaMemcpy(h_RGB_Temp, d_RGB_Final, (width * height * sizeof(float4)), cudaMemcpyDeviceToHost);
	
	//total sum of all the pixel values of the image
	float4 totalSum = h_RGB_Temp[ width * height - 1];


	//converts float r g b a values of the image integral sum into values between 0 and 1  by divding each pixel value with total sum and
	//converts these values to unsigned int values for visualizing the image.
	d_assign_final_image<<<gridSize, blockSize>>>(d_RGB_Final, d_dest, width, height, totalSum);
	


    // sync host and stop computation
    checkCudaErrors(cudaDeviceSynchronize());

	*/

////-------------------------------------------------------------------------------------------------------------------------------------------------------/////


}

#endif
